'''Utilities for Bash HTTP request generators.'''

from http_request_codegen.generators.python._utils import (  # noqa: F401
    escape_by_quote
)


DEFAULT_INDENT = '    '
DEFAULT_QUOTE_CHAR = '"'
DEFAULT_WRAP = 80
